/*
 * Copyright SpringSource Inc 2007,2009.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.springsource.util.common;

import java.util.Collection;
import java.util.Map;

/**
 * A set of useful assertions based on those provided by the Spring Framework's Assert class.
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * This class is thread safe.
 * 
 * @since Jersey
 * @see org.junit.Assert
 */
public final class Assert {

    /**
     * Assert a boolean expression, throwing a <code>IllegalArgumentException</code> if the test result is
     * <code>false</code>.
     * 
     * <pre class="code">
     * Assert.isTrue(i &gt; 0, &quot;The value must be greater than zero&quot;);
     * </pre>
     * 
     * @param expression a boolean expression
     * @param message the exception message to use if the assertion fails
     * @param inserts any inserts to include if the message is a format string.
     * @throws IllegalArgumentException if expression is <code>false</code>
     */
    public static void isTrue(boolean expression, String message, Object... inserts) {
    	if (!expression) {
    		throw new IllegalArgumentException(String.format(message, inserts));
    	}
    }
    
    /**
     * Assert a boolean expression, throwing a <code>IllegalArgumentException</code> if the test result is
     * <code>true</code>.
     * 
     * <pre class="code">
     * Assert.isFalse(state.isBroken(), &quot;The state is broken&quot;);
     * </pre>
     * 
     * @param expression a boolean expression
     * @param message the exception message to use if the assertion fails
     * @param inserts any inserts to include if the message is a format string.
     * @throws IllegalArgumentException if expression is <code>false</code>
     */
    public static void isFalse(boolean expression, String message, Object... inserts) {
        if (expression) {
        	throw new IllegalArgumentException(String.format(message, inserts));
        }
    }

    /**
     * Assert that an object is <code>null</code>.
     * 
     * <pre class="code">
     * Assert.isNull(value, &quot;The value must be null&quot;);
     * </pre>
     * 
     * @param object the object to check
     * @param message the exception message to use if the assertion fails
     * @param inserts any inserts to include if the message is a format string.
     * @throws IllegalArgumentException if the object is not <code>null</code>
     */
    public static void isNull(Object object, String message, Object... inserts) {
    	if (object != null) {
    		throw new IllegalArgumentException(String.format(message, inserts));
    	}
    }

    /**
     * Assert that an object is not <code>null</code>.
     * 
     * <pre class="code">
     * Assert.notNull(clazz, &quot;The class must not be null&quot;);
     * </pre>
     * 
     * @param object the object to check
     * @param message the exception message to use if the assertion fails
     * @param inserts any inserts to include if the message is a format string.
     * @throws IllegalArgumentException if the object is <code>null</code>
     */
    public static void notNull(Object object, String message, Object... inserts) {
        if (object == null) {
        	throw new IllegalArgumentException(String.format(message, inserts));
        }
    }

    /**
     * Assert that the given String is not empty; that is, it must not be <code>null</code> and not the empty String.
     * 
     * <pre class="code">
     * Assert.hasLength(name, &quot;Name must not be empty&quot;);
     * </pre>
     * 
     * @param text the String to check
     * @param message the exception message to use if the assertion fails
     * @param inserts any inserts to include if the message is a format string.
     * @see StringUtils#hasLength
     */
    public static void hasLength(String text, String message, Object... inserts) {
        if (text == null || text.length() == 0) {
        	throw new IllegalArgumentException(String.format(message, inserts));
        }
    }

    /**
     * Assert that an array has elements; that is, it must not be <code>null</code> and must have at least one
     * element.
     * 
     * <pre class="code">
     * Assert.notEmpty(array, &quot;The array must have elements&quot;);
     * </pre>
     * 
     * @param array the array to check
     * @param message the exception message to use if the assertion fails
     * @param inserts any inserts to include if the message is a format string.
     * @throws IllegalArgumentException if the object array is <code>null</code> or has no elements
     */
    public static void notEmpty(Object[] array, String message, Object... inserts) {
        if (array == null || array.length == 0) {
            throw new IllegalArgumentException(String.format(message, inserts));
        }
    }

    /**
     * Assert that a collection has elements; that is, it must not be <code>null</code> and must have at least one
     * element.
     * 
     * <pre class="code">
     * Assert.notEmpty(collection, &quot;Collection must have elements&quot;);
     * </pre>
     * @param <T> the type of the elements of the collection
     * 
     * @param collection the collection to check
     * @param message the exception message to use if the assertion fails
     * @param inserts any inserts to include if the message is a format string.
     * @throws IllegalArgumentException if the collection is <code>null</code> or has no elements
     */
    public static <T> void notEmpty(Collection<T> collection, String message, Object... inserts) {
    	if (collection == null || collection.isEmpty()) {
            throw new IllegalArgumentException(String.format(message, inserts));
        }
    }

    /**
     * Assert that a Map has entries; that is, it must not be <code>null</code> and must have at least one entry.
     * 
     * <pre class="code">
     * Assert.notEmpty(map, &quot;Map must have entries&quot;);
     * </pre>
     * @param <K> type of Key (domain)
     * @param <V> type of Value (range)
     * 
     * @param map the map to check
     * @param message the exception message to use if the assertion fails
     * @param inserts any inserts to include if the message is a format string.
     * @throws IllegalArgumentException if the map is <code>null</code> or has no entries
     */
    public static <K, V> void notEmpty(Map<K, V> map, String message, Object... inserts) {
        if (map == null || map.isEmpty()) {
            throw new IllegalArgumentException(String.format(message, inserts));
        }
    }

    /**
     * Assert that the provided object is a non-null instance of the provided class.
     * 
     * <pre class="code">
     * Assert.instanceOf(Foo.class, foo);
     * </pre>
     * @param <T> the type to check this for
     * 
     * @param type the type to check against
     * @param obj the object to check
     * @param message a message which will be prepended to the message produced by the function itself, and which may be
     *        used to provide context. It should normally end in a ": " or ". " so that the function generate message
     *        looks ok when appended to it.
     * @param inserts any inserts to include if the message is a format string.
     * @throws IllegalArgumentException if the object is not an instance of clazz
     * @see Class#isInstance
     */
    public static <T> void isInstanceOf(Class<T> type, Object obj, String message, Object... inserts) {
    	notNull(type, "The type to check against must not be null");
    	if (!type.isInstance(obj)) {
            throw new IllegalArgumentException(String.format(message, inserts) + "Object of class [" + (obj != null ? obj.getClass().getName() : "null")
                + "] must be an instance of " + type);
        }
    }

    /**
     * Assert that <code>superType.isAssignableFrom(subType)</code> is <code>true</code>.
     * 
     * <pre class="code">
     * Assert.isAssignable(Number.class, myClass);
     * </pre>
     * @param <T> the supertype
     * @param <U> the subtype
     * 
     * @param superType the super type to check against
     * @param subType the sub type to check
     * @param message a message which will be prepended to the message produced by the function itself, and which may be
     *        used to provide context. It should normally end in a ": " or ". " so that the function generate message
     *        looks ok when appended to it.
     * @param inserts any inserts to include if the message is a format string.
     * @throws IllegalArgumentException if the classes are not assignable
     */
    public static <T, U> void isAssignable(Class<T> superType, Class<U> subType, String message, Object... inserts) {
    	notNull(superType, "Type to check against must not be null");
    	if (subType == null || !superType.isAssignableFrom(subType)) {
    		throw new IllegalArgumentException(String.format(message, inserts) + subType + " is not assignable to " + superType);
        }
    }

    /*
     * Prevent instantiation - Java does not allow final abstract classes.
     */
    private Assert() {
    }
}
