/*
 * Copyright SpringSource Inc 2007,2009.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.springsource.util.common;

import java.util.Enumeration;
import java.util.Iterator;

/**
 * An <code>IterableEnumeration</code> wraps an {@link Enumeration} in an
 * implementation of {@link Iterable} thereby allowing the <code>Enumeration</code>
 * to be used in a for-each loop.
 * <p/>
 * 
 * <pre>
 * Enumeration<Foo> enumeration = ...
 * IterableEnumeration<Foo> foos = new IterableEnumeration<Foo>(enumeration);
 * 
 * for (Foo foo : foos) {
 *     // Do something with foo
 * }
 * </pre>
 * 
 * <p/>
 * <strong>Note:</strong> as <code>Enumerations</code> do not support entry removal
 * the {@link Iterator} returned by {@link #iterator()} does not support the optional
 * {@link Iterator#remove() remove} method: invocation of the method will result in an
 * {@link UnsupportedOperationException} being thrown.
 * 
 * <strong>Concurrent Semantics</strong><br />
 * This class is <strong>not</strong> thread-safe.
 * @param <T> the element type of the enumeration
 * 
 * @since Jersey
 */
public class IterableEnumeration<T> implements Iterable<T> {
    
    private final Iterator<T> iterator;
    
    /**
     * Creates a new <code>IterableEnumeration</code> backed by the supplied <code>Enumeration</code>
     * @param enumeration The <code>Enumeration</code> to be made iterable.
     */
    public IterableEnumeration(Enumeration<T> enumeration) {
        this.iterator = new EnumerationIterator<T>(enumeration);
    }

    /**
     * Returns the <code>Enumeration</code>-backed <code>Iterator</code>.
     * <p/>
     * <strong>Note:</strong> as <code>Enumerations</code> do not support entry removal
     * the returned {@link Iterator} does not support the optional
     * {@link Iterator#remove() remove} method: invocation of the method will result in an
     * {@link UnsupportedOperationException} being thrown.
     */
    public Iterator<T> iterator() {
        return iterator;
    }
    
    private class EnumerationIterator<E> implements Iterator<E> {
        
        private final Enumeration<E> enumeration;
        
        private EnumerationIterator(Enumeration<E> enumeration) {
            this.enumeration = enumeration;
        }

        public boolean hasNext() {
            return this.enumeration.hasMoreElements();
        }

        public E next() {
            return this.enumeration.nextElement();
        }

        public void remove() {
            throw new UnsupportedOperationException();
            
        }       
    }
}
