/*
 * Copyright SpringSource Inc 2009.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.springsource.util.common;

import java.util.Iterator;

/**
 * {@link SynchronizedIterator} wraps a given {@link Iterator} and protects access to the iterator by synchronizing on
 * the monitor of the containing {@link SynchronizedCollection}.
 * <p />
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * This class is thread safe, but ConcurrentModificationException may still be thrown.
 * 
 * @author Glyn Normington
 * @param <T> the type of element the iterator returns
 * @since jersey
 */
public class SynchronizedIterator<T> extends SynchronizedObject implements Iterator<T> {

    private final Object monitor;

    private final Iterator<T> iterator;

    public SynchronizedIterator(Iterator<T> iterator, Object monitor) {
        super(iterator, monitor);
        this.monitor = monitor;
        this.iterator = iterator;
    }

    /**
     * {@inheritDoc}
     */
    public boolean hasNext() {
        synchronized (this.monitor) {
            return this.iterator.hasNext();
        }
    }

    /**
     * {@inheritDoc}
     */
    public T next() {
        synchronized (this.monitor) {
            return this.iterator.next();
        }
    }

    /**
     * {@inheritDoc}
     */
    public void remove() {
        synchronized (this.monitor) {
            this.iterator.remove();
        }
    }

}