/*
 * Copyright SpringSource Inc 2009.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.springsource.util.common;

import java.util.ListIterator;

/**
 * {@link SynchronizedListIterator} wraps a given {@link ListIterator} and protects access to the iterator by
 * synchronizing on the monitor of the containing {@link SynchronizedList}.
 * <p />
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * This class is thread safe, but ConcurrentModificationException may still be thrown.
 * 
 * @author Glyn Normington
 * @param <T> the type of the list elements
 * @since jersey
 */
public class SynchronizedListIterator<T> extends SynchronizedIterator<T> implements ListIterator<T> {

    private final Object monitor;

    private final ListIterator<T> listIterator;

    public SynchronizedListIterator(ListIterator<T> listIterator, Object monitor) {
        super(listIterator, monitor);
        this.monitor = monitor;
        this.listIterator = listIterator;
    }

    /**
     * {@inheritDoc}
     */
    public void add(T e) {
        synchronized (this.monitor) {
            this.listIterator.add(e);
        }
    }

    /**
     * {@inheritDoc}
     */
    public boolean hasPrevious() {
        synchronized (this.monitor) {
            return this.listIterator.hasPrevious();
        }
    }

    /**
     * {@inheritDoc}
     */
    public int nextIndex() {
        synchronized (this.monitor) {
            return this.listIterator.nextIndex();
        }
    }

    /**
     * {@inheritDoc}
     */
    public T previous() {
        synchronized (this.monitor) {
            return this.listIterator.previous();
        }
    }

    /**
     * {@inheritDoc}
     */
    public int previousIndex() {
        synchronized (this.monitor) {
            return this.listIterator.previousIndex();
        }
    }

    /**
     * {@inheritDoc}
     */
    public void set(T e) {
        synchronized (this.monitor) {
            this.listIterator.set(e);
        }
    }

}