/*
 * Copyright SpringSource Inc 2009.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.springsource.util.common;

/**
 * {@link SynchronizedObject} wraps a given {@link Object} and protects access to the object by synchronizing on a given
 * monitor.
 * <p />
 * Note that <code>hashCode</code> and <code>equals</code> are not delegated as this object is distinct from the wrapped object.
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * This class is thread safe.
 * 
 * @author Glyn Normington
 * @since jersey
 */
public class SynchronizedObject extends Object {

    private final Object monitor;

    private final Object object;

    public SynchronizedObject(Object object, Object monitor) {
        if (object == null) {
            throw new IllegalArgumentException("null object");
        }
        if (monitor == null) {
            throw new IllegalArgumentException("null monitor");
        }
        this.object = object;
        this.monitor = monitor;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public String toString() {
        synchronized (this.monitor) {
            return this.object.toString();
        }
    }

}